"""A module for pre/postprocessing dates
"""

import numpy as np
import pandas as pd
import random
import pandas_flavor as pf
import warnings
from collections import defaultdict
from scipy.stats import truncnorm

from sklearn.utils.validation import check_array, FLOAT_DTYPES
from sklearn.preprocessing import KBinsDiscretizer, OrdinalEncoder
from sklearn.base import BaseEstimator, TransformerMixin

from synthesis.transformers._base import BaseReversibleTransformer
from synthesis.synthesizers.utils import dp_marginal_distribution, _normalize_distribution
from synthesis.evaluation import visual

@pf.register_dataframe_method
def bin_numeric_column(df: pd.DataFrame, column_name: str, n_bins: int, col_min: int, col_max: int, strategy: str = 'uniform'):
    """Bin a numeric column according to the number of bins. Can either use staregy 'uniform'
    for obtaining equal-width bins, or strategy 'quanitle' for bins with equal number of
    observations

    Preserves NaN values. Inspired by sklearn KBinsDiscretizer

    Manually add column min and max. As inference from the data can reveal individual values."""

    if strategy not in ['uniform', 'quantile']:
        raise ValueError("Strategy must either be 'uniform' or 'quantile'")

    df = df.copy()

    # exclude nan values
    mask_missing = df[column_name].isna()
    column = df.loc[~mask_missing, column_name][~mask_missing]

    # create equal-width bins
    if strategy == 'uniform':
        bin_edges = np.linspace(col_min, col_max, n_bins + 1)

    # create bins with equal number of observations
    elif strategy == 'quantile':
        quantiles = np.linspace(0, 100, n_bins + 1)
        column_values = np.append(column.values, [col_min, col_max])
        bin_edges = np.asarray(np.percentile(column_values, quantiles))

        # remove bins with too small width
        mask = np.ediff1d(bin_edges, to_begin=np.inf) > 1e-8
        bin_edges = bin_edges[mask]
        if len(bin_edges) - 1 != n_bins:
            warnings.warn('Bins whose width are too small (i.e., <= '
                          '1e-8). Consider decreasing the number of bins.')
            n_bins = len(bin_edges) - 1

    # Values which are close to a bin edge are susceptible to numeric instability
    # Add eps to X so these values are binned correctly with respect to their decimal truncation
    # eps = 1.e-8 + 1.e-5 * np.abs(column)
    # column_name = np.digitize(column_name + eps, bin_edges[1:])
    column = np.digitize(column, bin_edges[1:])

    column = np.clip(column, 0, n_bins - 1)

    # round to 2 decimals to prevent long numbers
    lower_bound = np.round(bin_edges[np.int_(column)], 2)
    upper_bound = np.round(bin_edges[np.int_(column) + 1], 2)

    # transform column to IntervalArray
    df.loc[~mask_missing, column_name] = pd.arrays.IntervalArray.from_arrays(lower_bound, upper_bound, closed='left')
    # force Interval dtype for whole column_name
    df[column_name] = pd.arrays.IntervalArray(df[column_name])
    return df


@pf.register_dataframe_method
def sample_from_binned_column(df: pd.DataFrame, column_name: str, numeric_type='int', mean=None, std=None):
    """Reverse the binning of a numeric column, by sampling from the bin ranges. If mean and standard
    deviation of column are given, then sample from truncated normal, else take the uniform distribution"""
    if numeric_type not in ['int', 'float']:
        raise ValueError("Numeric type must be 'int' or 'float'")
    df = df.copy()
    mask_missing = df[column_name].isna()

    try:
        column = pd.arrays.IntervalArray(df[column_name][~mask_missing])
    except TypeError:
        raise ValueError("Numeric column needs to be converted to bins first. "
                         "First run .bin_numeric_column on a continuous column")

    lower_bound, upper_bound = column.left, column.right

    # sample from truncated normal distribution
    if mean and std:
        low = (lower_bound - mean) / std
        upp = (upper_bound - mean) / std
        X = truncnorm(low, upp, loc=mean, scale=std)

        sampled_values = X.rvs()
        if numeric_type == 'int':
            sampled_values = np.around(sampled_values, decimals=0).astype(int)
        df.loc[~mask_missing, column_name] = sampled_values

    # sample from uniform distribution
    else:
        if numeric_type == 'int':
            df.loc[~mask_missing, column_name] = np.random.randint(lower_bound, upper_bound)
        elif numeric_type == 'float':
            df.loc[~mask_missing, column_name] = np.random.uniform(lower_bound, upper_bound)
    return df

@pf.register_dataframe_method
def sample_from_reversed_dict(df: pd.DataFrame, **generalize_dict):
    """reverse generalization schema and sample unique values.
    dict structure: {column_name: {value: generalized_value}}"""
    for column_name, mapper in generalize_dict.items():
        df = _sample_from_reversed_dict(df, column_name, mapper)
    return df

def _sample_from_reversed_dict(df: pd.DataFrame, column_name, mapper):

    df = df.copy()

    # reverse schema
    reversed_dict = defaultdict(list)
    for key, value in mapper.items():
        reversed_dict[value].append(key)

    # sample from values
    column_generalized = df[column_name].values
    column_sampled = np.empty_like(column_generalized)
    for i in range(len(column_sampled)):
        reverse_candidates = reversed_dict[column_generalized[i]]

        if reverse_candidates:
            column_sampled[i] = random.choice(reverse_candidates)
        else:
            column_sampled[i] = column_generalized[i]
    df[column_name] = column_sampled

    # convert string 'nan' to NaN which automatically converts column to float if all other elements are numeric
    df[column_name] = df[column_name].replace({'nan': np.nan})
    return df

@pf.register_dataframe_method
def replace_rare_values(df, column, min_support=0.01, new_name='other'):
    """Replace values that occur less than minimum support with new name"""
    df = df.copy()

    # normalize if floating number, use counts if int
    if isinstance(min_support, float):
        normalize = True
        n_records_support = np.ceil(min_support * df.shape[0])
    else:
        normalize = False
        n_records_support = min_support

    column_value_counts = df[column].value_counts(normalize=normalize, dropna=False)

    # if no rare categories under min support - no change required
    if not (column_value_counts < min_support).any():
        return df

    rare_categories = list(column_value_counts[column_value_counts < min_support].index)

    # if mode then replace rare categories with the most frequent value in column (including NaN)
    if new_name == 'mode':
        new_name = df[column].mode(dropna=False)[0]

    print(f"Categories in column '{column}' that occur less than {n_records_support}x "
          f"and are replaced by {new_name} - converted categories: {rare_categories}")

    map_value_probs_to_rows = df[column].map(column_value_counts)
    convert_low_freq_rows = df[column].mask(map_value_probs_to_rows < min_support, new_name)
    df[column] = convert_low_freq_rows
    return df

class GeneralizeContinuous(KBinsDiscretizer):

    def __init__(self, n_bins=10, strategy='uniform', labeled_missing=None):
        super().__init__(n_bins=n_bins, strategy=strategy, encode='ordinal')
        # self.n_bins = n_bins
        # self.strategy = strategy

        self.labeled_missing = labeled_missing

    def fit(self, X, y=None):
        """
        Fit the estimator.
        Parameters
        ----------
        X : numeric array-like, shape (n_samples, n_features)
            Data to be discretized.
        y : None
            Ignored. This parameter exists only for compatibility with
            :class:`sklearn.pipeline.Pipeline`.
        Returns
        -------
        self
        """
        self._infer_numerical_type(X)
        self._header = X.columns

        # X = self._validate_data(X, dtype='numeric')
        X = check_array(X, dtype='numeric', force_all_finite='allow-nan')

        valid_strategy = ('uniform', 'quantile')
        if self.strategy not in valid_strategy:
            raise ValueError("Valid options for 'strategy' are {}. "
                             "Got strategy={!r} instead."
                             .format(valid_strategy, self.strategy))

        n_features = X.shape[1]
        n_bins = self._validate_n_bins(n_features)

        bin_edges = np.zeros(n_features, dtype=object)
        for jj in range(n_features):
            column = X[:, jj]
            missing_idx = self._get_missing_idx(column)
            column = column[~missing_idx]
            col_min, col_max = column.min(), column.max()

            if col_min == col_max:
                warnings.warn("Feature %d is constant and will be "
                              "replaced with 0." % jj)
                n_bins[jj] = 1
                bin_edges[jj] = np.array([-np.inf, np.inf])
                continue

            if self.strategy == 'uniform':
                bin_edges[jj] = np.linspace(col_min, col_max, n_bins[jj] + 1)

            elif self.strategy == 'quantile':
                quantiles = np.linspace(0, 100, n_bins[jj] + 1)
                bin_edges[jj] = np.asarray(np.percentile(column, quantiles))


            # Remove bins whose width are too small (i.e., <= 1e-8)
            if self.strategy == 'quantile':
                mask = np.ediff1d(bin_edges[jj], to_begin=np.inf) > 1e-8
                bin_edges[jj] = bin_edges[jj][mask]
                if len(bin_edges[jj]) - 1 != n_bins[jj]:
                    warnings.warn('Bins whose width are too small (i.e., <= '
                                  '1e-8) in feature %d are removed. Consider '
                                  'decreasing the number of bins.' % jj)
                    n_bins[jj] = len(bin_edges[jj]) - 1

        self.bin_edges_ = bin_edges
        self.n_bins_ = n_bins

        return self

    def transform(self, X):
        """
        Discretize the data.
        Parameters
        ----------
        X : numeric array-like, shape (n_samples, n_features)
            Data to be discretized.
        Returns
        -------
        Xt : numeric array-like or sparse matrix
            Data in the binned space.
        """
        # check_is_fitted(self, attributes=None)

        Xt = check_array(X, copy=True, dtype=FLOAT_DTYPES, force_all_finite='allow-nan')
        n_features = self.n_bins_.shape[0]
        if Xt.shape[1] != n_features:
            raise ValueError("Incorrect number of features. Expecting {}, "
                             "received {}.".format(n_features, Xt.shape[1]))

        bin_edges = self.bin_edges_
        for jj in range(Xt.shape[1]):
            missing_idx = self._get_missing_idx(Xt[:, jj])

            # Values which are close to a bin edge are susceptible to numeric
            # instability. Add eps to X so these values are binned correctly
            # with respect to their decimal truncation. See documentation of
            # numpy.isclose for an explanation of ``rtol`` and ``atol``.
            rtol = 1.e-5
            atol = 1.e-8
            eps = atol + rtol * np.abs(Xt[~missing_idx, jj])
            Xt[~missing_idx, jj] = np.digitize(Xt[~missing_idx, jj] + eps, bin_edges[jj][1:])
        np.clip(Xt, 0, self.n_bins_ - 1, out=Xt)
        return pd.DataFrame(Xt, columns=self._header)

    def _infer_numerical_type(self, X):
        """Determine if numerical column is an integer of float for inverse transform"""
        # assert X.select_dtypes(exclude=['int', 'float']).shape[1] == 0, "input X contains non-numeric columns"
        self.integer_columns = []
        self.float_columns = []

        for c in X.columns:
            only_integers = (X[c].dropna().astype(float) % 1 == 0).all()
            if only_integers:
                self.integer_columns.append(c)
            else:
                self.float_columns.append(c)
            # conversion from string to int directly can cause ValueError
            # if np.array_equal(X[c].dropna(), X[c].dropna().astype(float).astype(int)):
            #     self.integer_columns.append(c)
            # else:
            #     self.float_columns.append(c)

    def _get_missing_idx(self, column):
        return np.isnan(column) | np.isin(column, self.labeled_missing)


    def inverse_transform(self, Xt):
        # check_is_fitted(self)
        assert set(Xt.columns) == set(self._header), "input contains different columns than seen in fit"

        Xinv = check_array(Xt, copy=True, dtype=FLOAT_DTYPES, force_all_finite='allow-nan')
        # Xinv = Xt.copy()
        n_features = self.n_bins_.shape[0]
        if Xinv.shape[1] != n_features:
            raise ValueError("Incorrect number of features. Expecting {}, "
                             "received {}.".format(n_features, Xinv.shape[1]))

        for jj in range(n_features):
            bin_edges = self.bin_edges_[jj]
            missing_idx = self._get_missing_idx(Xinv[:, jj])

            lower_bounds = bin_edges[np.int_(Xinv[~missing_idx, jj])]
            upper_bounds = bin_edges[np.int_(Xinv[~missing_idx, jj]) + 1]
            Xinv[~missing_idx, jj] = np.random.uniform(lower_bounds, upper_bounds)

            # todo transfer to numpy
            if self._header[jj] in self.integer_columns:
                Xinv[~missing_idx, jj] = np.round(Xinv[~missing_idx, jj])

        return pd.DataFrame(Xinv, columns=self._header)



class GeneralizeCategorical(GeneralizeContinuous):

    def __init__(self, epsilon=1.0, n_bins=5, strategy='uniform', max_cardinality=10):
        super().__init__(n_bins=n_bins, strategy=strategy)
        self.epsilon = epsilon
        self.max_cardinality = max_cardinality


    def fit(self, X, y=None):
        """ Steps:
        1. Transform categorical to continuous
        2. Store DP marginal counts for optional inverse transform
        3. Run super().fit() to get groups
        """
        X = X.astype(str).fillna('missing')
        self._ordinalencoder = OrdinalEncoder().fit(X)
        #todo: turn into numpy -> df needed for marginal distribution
        X_enc = self._ordinalencoder.transform(X)
        X_enc = pd.DataFrame(X_enc, columns=X.columns)

        # get dp marginal of encoded feature
        # todo turn into list of arrays
        local_epsilon = self.epsilon / X.shape[1]
        self.marginals_ = []
        for jj, c in enumerate(X.columns):
            self.marginals_.append(dp_marginal_distribution(X_enc.loc[:, c], local_epsilon).values)

        return super().fit(X_enc, y)

    def transform(self, X):
        """Equivalent to continuous transform but we still need to encode the data beforehand"""
        X = X.astype(str).fillna('missing')
        X_enc = self._ordinalencoder.transform(X)
        return super().transform(X_enc)

    def inverse_transform(self, Xt):
        assert set(Xt.columns) == set(self._header), "input contains different columns than seen in fit"

        X_enc = check_array(Xt, copy=True, dtype=FLOAT_DTYPES, force_all_finite='allow-nan')
        # Xinv = Xt.copy()
        n_records, n_features = X_enc.shape
        if X_enc.shape[1] != n_features:
            raise ValueError("Incorrect number of features. Expecting {}, "
                             "received {}.".format(n_features, X_enc.shape[1]))


        self._marginal_group_alloc = []

        for jj, c in enumerate(Xt.columns):
            bin_edges = self.bin_edges_[jj]
            marginals = self.marginals_[jj]
            marginals_idx = np.arange(len(marginals))

            rtol = 1.e-5
            atol = 1.e-8
            eps = atol + rtol * np.abs(marginals)
            marginal_group_alloc = np.digitize(marginals_idx + eps, bin_edges[1:])
            np.clip(marginal_group_alloc, 0, self.n_bins_[jj] - 1, out=marginal_group_alloc)

            self._marginal_group_alloc.append(marginal_group_alloc)

            # lower_bounds = np.int_(bin_edges[np.int_(X_enc[:, jj])])
            # upper_bounds = np.int_(bin_edges[np.int_(X_enc[:, jj]) + 1])

            for i in range(n_records):
                # Values which are close to a bin edge are susceptible to numeric
                # instability. Add eps to X so these values are binned correctly
                # with respect to their decimal truncation. See documentation of
                # numpy.isclose for an explanation of ``rtol`` and ``atol``.
                # rtol = 1.e-5
                # atol = 1.e-8
                # eps = atol + rtol * np.abs(upper_bounds[i])
                # marginal_candidates = marginals[
                #     (marginals.keys() >= lower_bounds[i]) &
                #     (marginals.keys() < upper_bounds[i] + eps)]

                #np.where returns 1d tuple, thus index 0
                marginal_candidate_idx = np.where(X_enc[i, jj] == marginal_group_alloc)[0]
                marginal_candidate_probs = marginals[marginal_candidate_idx]
                marginal_candidate_probs_normalized = _normalize_distribution(marginal_candidate_probs)


                # marginal_idx = np.arange(lower_bounds[i], upper_bounds[i])
                # marginal_probs = marginals[marginal_idx]

                # marginal_probs_normalized = marginal_probs / marginal_probs.sum()
                # sample encoded (numerical) value based on marginal probabilities
                # print(jj)
                # print(X_enc.shape)
                # print(X_enc[i, jj])
                # print(marginal_candidate_idx)
                X_enc[i, jj] = np.random.choice(marginal_candidate_idx, p=marginal_candidate_probs_normalized)

                # X_enc[i, jj] = np.random.choice(list(marginal_candidates.keys()), p=marginal_candidates_normalized.values)

        # inverse transform numerical value to original categorical
        X_inv = self._ordinalencoder.inverse_transform(X_enc)
        return pd.DataFrame(X_inv, columns=self._header)



def get_high_cardinality_features(X, threshold=50):
    """Get features with more unique values than the specified threshold."""
    high_cardinality_features = []
    for c in X.columns:
        if X[c].nunique() > threshold:
            high_cardinality_features.append(c)
    return high_cardinality_features


class GeneralizeSchematic(TransformerMixin, BaseEstimator):

    def __init__(self, schema_dict, label_unknown=None):
        self.schema_dict = schema_dict
        self.label_unknown = label_unknown

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        """Replaces all values in col with its generalized form in schema dict"""
        Xt = X.copy()

        # convert categories not present in schema to value given for label_unknown
        if self.label_unknown:
            Xt[~Xt.isin(self.schema_dict.keys())] = self.label_unknown

        return Xt.replace(self.schema_dict)

    def _reverse_schema(self):
        """reverse schema dict with non-unique values"""
        reversed_dict = defaultdict(list)

        for key, value in self.schema_dict.items():
            reversed_dict[value].append(key)
        return reversed_dict

    def inverse_transform(self, X):
        """reverse the schema by sampling from the available candidates"""
        reversed_schema = self._reverse_schema()
        X_generalized = X.values
        X_sampled = np.empty_like(X)
        for i in range(len(X)):
            reverse_candidates = reversed_schema[X_generalized[i]]

            if reverse_candidates:
                X_sampled[i] = np.random.choice(reverse_candidates)
            else:
                X_sampled[i] = X_generalized[i]
        return X_sampled


class GroupRareCategories(BaseReversibleTransformer):
    """Transformer to group rare categories"""

    def __init__(self, threshold=0.05, name_group='Other'):
        self.threshold = threshold
        self.name_group = name_group

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        """Group categories that occur less then threshold"""
        return X.mask(X.map(X.value_counts(normalize=True)) < self.threshold, self.name_group)

# class GeneralizeCategorical(GeneralizeContinuous):
#
#     def __init__(self, epsilon=1.0, n_bins=10, strategy='uniform', labeled_missing=None):
#         super().__init__(n_bins=n_bins, strategy=strategy, encode='ordinal')
#         self.epsilon = epsilon
#
#     def fit(self, X, y=None):
#         # if user-specified bins in form of iterable or dict
#         if self.bins:
#             pass
#
#         self.marginals_ = {}
#         self._ordinalencoders = {}
#         X_enc = np.empty_like(X)
#         for jj, c in enumerate(X.columns):
#             uniques = sorted(set(X[c]))
#             uniques, counts = X[c].value_counts(dropna=False)
#
#             local_epsilon = self.epsilon / X.shape[1]
#             self.marginals_[c] = dp_marginal_distribution(X[c], local_epsilon)
#
#             # get numeric values - store encoder for inverse transform
#             self._ordinalencoders[c] = OrdinalEncoder().fit(X[c])
#             X_enc[:, jj] = self._ordinalencoders[c].transform(X[c])
#
#
#         return super().fit(X_enc, y)
#
#     def inverse_transform(self, Xt):
#         X_enc = super().inverse_transform(Xt)
#         for jj in range(X_enc.shape[1]):
#             # todo fix column names indexing
#             # todo ensure inverse gives back X_enc which OrdinalEncoder.inverse_transform(X_le)
#             pass




"""Classes for transforming sequential data
"""

import numpy as np
import pandas as pd
import datetime
from pathlib import Path


from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_random_state


class FlattenSequentialRecords(TransformerMixin, BaseEstimator):
    pass


class GeneralizeDateSequence(TransformerMixin, BaseEstimator):

    def __init__(self, date_sequence=None):
        self.date_sequence = date_sequence

    def fit(self, X, y=None):
        """A reference implementation of a fitting function for a transformer.
        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The training input samples. Contain only date columns
        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.
        Returns
        -------
        self : object
            Returns self.
        """
        if self.date_sequence is None:
            self.date_sequence = X.columns
            print("no data_sequence specified in init - assume columns in X "
                  "represent the ordered data sequence")
        return self

    def transform(self, X, y=None):
        X = X.copy()
        self.reference_date_col = self.date_sequence[0]

        # for col in self.date_sequence:
        #     try:
        #         X[col] = pd.to_datetime(X[col], infer_datetime_format=True)
        #     except:
        #         X[col] = pd.to_datetime(X[col], infer_datetime_format=True, errors='coerce')
        #
        X = pd.to_datetime(X.stack()).unstack()
        reverse_date_sequence = self.date_sequence[::-1]

        for idx, col in enumerate(reverse_date_sequence):
            if col is not self.reference_date_col:
                prior_event_date = X[reverse_date_sequence[idx+1]]
                days_after_prior_event = (X[col] - prior_event_date).dt.days
                X[col] = days_after_prior_event

        # cat_year_month = X[reference_date_col].astype(str).str.rsplit('-', 1).str[0]
        cat_year_month = X[self.reference_date_col].dt.to_period('m')
        X[self.reference_date_col] = cat_year_month
            # X[c + '_year'] = X[c].dt.year
            # X[c + '_month'] = X[c].dt.month
            # del X[c]
        return X

    def inverse_transform(self, Xt):
        """Transform categorical dates back to datetime format"""

        assert Xt.columns.isin(self.date_sequence).all(), 'input dataframe contains columns not seen in fit'

        if not pd.core.dtypes.common.is_period_dtype(Xt[self.reference_date_col]):
            Xt[self.reference_date_col] = pd.to_datetime(Xt[self.reference_date_col].astype(str)).dt.to_period('m')

        reference_period_ym = Xt[self.reference_date_col]
        days_in_month = reference_period_ym.dt.days_in_month

        sample_days = np.random.uniform(0, days_in_month)
        # missing dates have days in month of -1
        sample_days[sample_days < 0] = np.nan
        sample_days = np.ceil(sample_days)

        reference_date_ymd = pd.to_datetime({'year': reference_period_ym.dt.year,
                                             'month': reference_period_ym.dt.month,
                                             'day': sample_days})

        # Xinv = check_array(Xt, copy=True, dtype=FLOAT_DTYPES)
        Xinv = Xt.copy()
        Xinv[self.reference_date_col] = reference_date_ymd

        for c in self.date_sequence[1:]:
            # add days to reference date
            Xinv[c] = reference_date_ymd + pd.to_timedelta(Xinv[c].astype(float), unit='d')
            reference_date_ymd = Xinv[c]

        return Xinv


if __name__ == '__main__':
    data_path = Path("c:/data/1_iknl/processed/crc_behandeling.csv")
    X = pd.read_csv(data_path)
    columns = ['gbs_begin_dtm', 'gbs_eind_dtm']
    X = X.loc[:, columns]
    print(X.head(20))

    date_sequence = columns
    gen_dates = GeneralizeDateSequence(date_sequence)
    X_cat = gen_dates.fit_transform(X)
    print(X_cat.head(20))

    X_inv = gen_dates.inverse_transform(X_cat)
    print(X_inv.head(20))


"""Module with functions for visualizing the difference between datasets"""
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from lifelines import KaplanMeierFitter

from synthesis.evaluation import metrics


def plot_kmf_comparison(datasets, dataset_names, T_varname, E_varname, G_varname):
    """
    Plot side-bys-side kaplan-meier of input datasets

    Parameters
    ----------
    datasets: list of input data
    dataset_names: names of input data - note: len equal to datasets
    T_varname: time variable name
    E_varname: event variable name
    G_varname: grouping variable name

    Returns Kaplan-Meier plot of input datasets
    -------

    """
    if not isinstance(datasets, list):
        datasets = [datasets]
    if not isinstance(dataset_names, list):
        dataset_names = [dataset_names]
    assert len(datasets) == len(dataset_names), "input datasets and dataset_names are of different lengths"

    figsize = (8 * len(datasets), 7)
    fig, ax = plt.subplots(1, len(datasets), figsize=figsize, sharey=True)

    sns.set(font_scale=1.5)
    #     sns.set_context('paper', rc={"lines.linewidth": 1.2})
    sns.despine()
    palette = ['#0d3d56', '#006887', '#0098b5', '#00cbde', '#00ffff']

    for X, X_name, ax_cur in zip(datasets, dataset_names, ax):
        T = X[T_varname].astype(float)
        E = X[E_varname].astype(float)

        kmf = KaplanMeierFitter()
        unique_values = np.sort(X[G_varname].unique())
        for g, color in zip(unique_values, palette):
            mask = (X[G_varname] == g)
            kmf.fit(T[mask], event_observed=E[mask], label=g)
            ax_cur = kmf.plot(ax=ax_cur, color=color)
            ax_cur.legend(title=G_varname)
            ax_cur.set_title('Survival Analysis C50 - {} Data'.format(X_name))
            ax_cur.set_ylim(0, 1)
    plt.tight_layout()



"""Base classes for all synthesizers"""

import numpy as np
import pandas as pd

from sklearn.base import BaseEstimator, TransformerMixin

class BaseReversibleTransformer(TransformerMixin, BaseEstimator):

    def __init__(self):
        super().__init__()

    def fit(self, X, y=None):
        X = self._check_input_data(X)
        return self

    def transform(self, X, y=None):
        raise NotImplementedError("Implement transform method")

    def inverse_transform(self, Xt):
        Xt = self._check_output_data(Xt)
        return Xt

    def _check_input_data(self, data_input):
        self.dtypes_fit_ = data_input.dtypes
        return data_input

    def _check_output_data(self, data_output):
        # convert dtypes back to original - first convert object to bool to prevent all values from becoming True
        bool_cols = self.dtypes_fit_[self.dtypes_fit_ == bool].index
        data_output[bool_cols] = data_output[bool_cols].replace({'False': False, 'True': True})
        data_output = data_output.astype(self.dtypes_fit_)
        return data_output


"""Base classes for all synthesizers"""

import numpy as np
import pandas as pd
import dill

from abc import ABC, abstractmethod
from copy import copy
from numbers import Real
from scipy.spatial.distance import jensenshannon

from synthesis.synthesizers.utils import astype_categorical

class BaseDPSynthesizer(ABC):
    """Abstract base class for all differentially private synthesizers"""

    def __init__(self, epsilon=1.0, verbose=True):
        """Base class for differentially private synthesizers.

        Parameters
        ----------
        epsilon : float or int
            Privacy parameter epsilon in differential privacy. Must be in range [0, float(np.inf)].
        verbose : bool
            Enable verbose output
        """
        self.epsilon = epsilon
        self.verbose = verbose

    @abstractmethod
    def fit(self, data):
        """Fit synthesizer to input data.

         Parameters
        ----------
        data: pandas.DataFrame
            Input dataset to fit the synthesizer to.

        Returns
        -------
        self : class
            Returns the fitted synthesizer that can be sampled from.
        """
        pass

    @abstractmethod
    def sample(self, n_records=None):
        """Sample records from fitted synthesizer.

        Parameters
        ----------
        n_records: int or float
            Number of records to be sampled.

        Returns
        -------
        pandas.DataFrame
            Returns synthetic dataset that mimics the input data seen in fit.
        """
        pass

    def score(self, original_data, synthetic_data, score_dict=False):
        """Calculate jensen_shannon distance between original and synthetic data.
        Look for more elaborate evaluation techniques in synthesis.evaluation.

        Parameters
        ----------
        original_data: pandas.DataFrame
            Original data that was seen in fit
        synthetic_data: pandas.DataFrame
            Synthetic data that was generated based original_data
        score_dict: bool
            If true, will return jensen_shannon scores of each column individually
        Returns
        -------
        average jensen_shannon distance: float
            Average jensen_shannon distance over all columns
        per-column jensen_shannon distance (if score_dict): dict
            Per-column jensen_shannon distance
        """
        column_distances = {}
        for i, c in enumerate(original_data.columns):
            # compute value_counts for both original and synthetic - align indexes as certain column
            # values in the original data may not have been sampled in the synthetic data
            counts_original, counts_synthetic = \
                original_data[c].value_counts(dropna=False).align(
                    synthetic_data[c].value_counts(dropna=False), join='outer', axis=0, fill_value=0
                )

            js_distance = jensenshannon(counts_original, counts_synthetic)
            column_distances[c] = js_distance
        average_column_distance = sum(column_distances.values()) / len(original_data.columns)
        if score_dict:
            return average_column_distance, column_distances
        return average_column_distance


    def copy(self):
        """Produces a copy of the class.
        Returns
        -------
        self : class
            Returns the copy.
        """
        return copy(self)

    def save(self, path):
        """
        Save this synthesizer instance to the given path using pickle.

        Parameters
        ----------
        path: str
            Path where the synthesizer instance is saved.
        """
        with open(path, 'wb') as output:
            dill.dump(self, output)

    @classmethod
    def load(cls, path):
        """Load a synthesizer instance from specified path.
        Parameters
        ----------
        path: str
            Path where the synthesizer instance is saved.

        Returns
        -------
        synthesizer : class
            Returns synthesizer instance.
        """
        with open(path, 'rb') as f:
            return dill.load(f)

    def _check_init_args(self):
        """Check arguments provided at object instantiation"""
        self._check_epsilon()

    def _check_input_data(self, data):
        """Check input dataset - save column names, number of records and column dtypes"""
        # prevent integer column indexing issues
        data.columns = data.columns.astype(str)

        # store general information about input data to ensure synthetic data maintains the same structure
        self.columns_ = list(data.columns)
        self.n_records_fit_ = data.shape[0]
        self.dtypes_fit_ = data.dtypes

        # converts dataframe to categorical, include 'nan' as category
        data = astype_categorical(data, include_nan=True)
        return data

    def _check_output_data(self, data_synth):
        """Check if output is in same format input data."""
        # DEPRCATED: bool conversion no longer special case as we change input_data.astype(str) to input_data.astype(category)
        # convert dtypes back to original - first convert object to bool to prevent all values from becoming True
        # bool_cols = self.dtypes_fit_[self.dtypes_fit_ == bool].index

        # data_synth[bool_cols] = data_synth[bool_cols].replace({'False': False, 'True': True})
        data_synth = data_synth.astype(self.dtypes_fit_)

        # convert 'nan' to NaN
        data_synth = data_synth.replace({'nan': np.nan})

        # convert column order as seen in fit
        data_synth = data_synth[self.columns_]
        return data_synth

    def _check_epsilon(self):
        """Check whether epsilon is in range [0, float(np.inf)]"""
        if not isinstance(self.epsilon, Real):
            raise TypeError("Epsilon must be numeric")

        if self.epsilon < 0:
            raise ValueError("Epsilon must be non-negative")

        if self.epsilon == 0:
            raise ValueError("Epsilon cannot be zero")
        float(self.epsilon)

    def _check_is_fitted(self):
        if not hasattr(self, 'model_'):
            msg = ("This %(name)s instance is not fitted yet. Call 'fit' with "
                   "appropriate arguments before using this synthesizer.")
            raise NotFittedError(msg % {'name': type(self).__name__})


class NotFittedError(Exception):
    """Exception to indicate that the synthesizer is not fitted yet"""



"""
Synthetic Data Generation via contingency tables.
"""

import numpy as np
import pandas as pd

from synthesis.synthesizers._base import BaseDPSynthesizer
from synthesis.synthesizers.utils import dp_contingency_table, cardinality


class ContingencySynthesizer(BaseDPSynthesizer):
    """Synthetic Data Generation via contingency table.

    Creates a contingency tables based on the whole input dataset.
    Note: memory-intensive - data with many columns and/or high column cardinality might not fit into memory.
    """

    def __init__(self, epsilon=1.0, verbose=True):
        super().__init__(epsilon=epsilon, verbose=verbose)

    def fit(self, data):
        data = self._check_input_data(data)
        self._check_init_args()

        ct_size = cardinality(data)
        print('Estimated size contingency table: {}'.format(ct_size))

        self.model_ = dp_contingency_table(data, self.epsilon)
        if self.verbose:
            print('Contingency table fitted')
        return self

    def sample(self, n_records=None):
        self._check_is_fitted()
        n_records = n_records or self.n_records_fit_

        prob = np.array(self.model_) / sum(self.model_)
        idx = np.arange(len(self.model_))

        sampled_idx = np.random.choice(idx, size=n_records, p=prob, replace=True)
        sampled_records = self.model_[sampled_idx].index.to_frame(index=False)
        return sampled_records



if __name__ == '__main__':
    data_path = 'C:/projects/synthetic_data_generation/examples/data/original/adult_8c.csv'
    df = pd.read_csv(data_path)
    epsilon = float(np.inf)
    df_sub = df[['education', 'occupation', 'relationship']]

    cs = ContingencySynthesizer(epsilon=epsilon)
    cs.fit(df_sub)
    df_cs = cs.sample()
    df_cs.head()

    cs.score(df_sub, df_cs, score_dict=True)




"""
General utility functions and building blocks for synthesizers
"""
import numpy as np
import pandas as pd

from diffprivlib.mechanisms import LaplaceTruncated
from thomas.core import Factor, CPT, JPT
from sys import maxsize

def dp_contingency_table(data, epsilon):
    """Compute differentially private contingency table of input data"""
    contingency_table_ = contingency_table(data)

    # if we remove one record from X the count in one cell decreases by 1 while the rest stays the same.
    sensitivity = 1
    dp_mech = LaplaceTruncated(epsilon=epsilon, lower=0, upper=maxsize, sensitivity=sensitivity)

    contingency_table_values = contingency_table_.values.flatten()
    dp_contingency_table = np.zeros_like(contingency_table_values)
    for i in np.arange(dp_contingency_table.shape[0]):
        # round counts upwards to preserve bins with noisy count between [0, 1]
        dp_contingency_table[i] = np.ceil(dp_mech.randomise(contingency_table_values[i]))

    return Factor(dp_contingency_table, states=contingency_table_.states)


def dp_marginal_distribution(data, epsilon):
    """Compute differentially private marginal distribution of input data"""
    marginal_ = marginal_distribution(data)

    # removing one record from X will decrease probability 1/n in one cell of the
    # marginal distribution and increase the probability 1/n in the remaining cells
    sensitivity = 2 / data.shape[0]
    dp_mech = LaplaceTruncated(epsilon=epsilon, lower=0, upper=maxsize, sensitivity=sensitivity)

    marginal_values = marginal_.values.flatten()
    dp_marginal = np.zeros_like(marginal_.values)

    for i in np.arange(dp_marginal.shape[0]):
        # round counts upwards to preserve bins with noisy count between [0, 1]
        dp_marginal[i] = dp_mech.randomise(marginal_.values[i])

    dp_marginal = _normalize_distribution(dp_marginal)
    return Factor(dp_marginal, states=marginal_.states)


def dp_joint_distribution(data, epsilon):
    """Compute differentially private joint distribution of input data"""
    joint_distribution_ = joint_distribution(data)

    # removing one record from X will decrease probability 1/n in one cell of the
    # joint distribution and increase the probability 1/n in the remaining cells
    sensitivity = 2 / data.shape[0]
    dp_mech = LaplaceTruncated(epsilon=epsilon, lower=0, upper=maxsize, sensitivity=sensitivity)

    joint_distribution_values = joint_distribution_.values.flatten()
    dp_joint_distribution_ = np.zeros_like(joint_distribution_values)

    for i in np.arange(dp_joint_distribution_.shape[0]):
        dp_joint_distribution_[i] = dp_mech.randomise(joint_distribution_values[i])

    dp_joint_distribution_ = _normalize_distribution(dp_joint_distribution_)
    return JPT(dp_joint_distribution_, states=joint_distribution_.states)


def dp_conditional_distribution(data, epsilon, conditioned=None):
    """Compute differentially private conditional distribution of input data
    Inferred from marginal or joint distribution"""
    # if only one columns (series or dataframe), i.e. no conditioning columns
    if len(data.squeeze().shape) == 1:
        dp_distribution = dp_marginal_distribution(data, epsilon=epsilon)
    else:
        dp_distribution = dp_joint_distribution(data, epsilon=epsilon)
    cpt = CPT(dp_distribution, conditioned=conditioned)

    # normalize if cpt has conditioning columns
    if cpt.conditioning:
        cpt = _normalize_cpt(cpt)
    return cpt


"""Non-differentially private functions below"""

def contingency_table(data):
    return Factor.from_data(data)

def joint_distribution(data):
    """Get joint distribution by normalizing contingency table"""
    return contingency_table(data).normalize()

def marginal_distribution(data):
    assert len(data.squeeze().shape) == 1, "data can only consist of a single column"
    # converts single column dataframe to series
    data = data.squeeze()

    marginal = data.value_counts(normalize=True, dropna=False)
    states = {data.name: marginal.index.tolist()}
    return Factor(marginal, states=states)

def uniform_distribution(data):
    assert len(data.squeeze().shape) == 1, "data can only consist of a single column"
    # converts single column dataframe to series
    data = data.squeeze()
    n_unique = data.nunique(dropna=False)
    uniform = np.full(n_unique, 1/n_unique)
    states = {data.name: data.unique().tolist()}
    return Factor(uniform, states=states)

def compute_distribution(data):
    """"Draws a marginal or joint distribution depending on the number of input dimensions"""
    if len(data.squeeze().shape) == 1:
        return marginal_distribution(data)
    else:
        return joint_distribution(data)

"""Check and fix distributions"""


def _normalize_distribution(distribution):
    """Check whether probability distribution sums to 1"""
    distribution = _check_all_zero(distribution)
    distribution = distribution / distribution.sum()
    return distribution


def _check_all_zero(distribution):
    """In case distribution contains only zero values due to DP noise, convert to uniform"""
    if not np.any(distribution):
        distribution = np.repeat(1 / len(distribution), repeats=len(distribution))
    return distribution


def _normalize_cpt(cpt):
    """normalization of cpt with option to fill missing values with uniform distribution"""
    # convert to series as normalize does not work with thomas cpts
    series = cpt.as_series()
    series_norm_full = series / series.unstack().sum(axis=1)
    # fill missing combinations with uniform distribution
    uniform_prob = 1 / len(cpt.states[cpt.conditioned[-1]])
    series_norm_full = series_norm_full.fillna(uniform_prob)
    return CPT(series_norm_full, cpt.states)

def _ensure_arg_is_list(arg):
    if not arg:
        raise ValueError('Argument is empty: {}'.format(arg))

    arg = [arg] if isinstance(arg, str) else arg
    arg = list(arg) if isinstance(arg, tuple) else arg
    assert isinstance(arg, list), "input argument should be either string, tuple or list"
    return arg

def cardinality(X):
    """Compute cardinality of input data"""
    return np.prod(X.nunique(dropna=False))

def rank_columns_on_cardinality(X):
    """Rank columns based on number of unique values"""
    return X.nunique().sort_values(ascending=False)

def astype_categorical(data, include_nan=True):
    """Convert data to categorical and optionally adds nan as unique category"""
    # converts to dataframe in case of numpy input and make all columns categorical.
    data = pd.DataFrame(data).astype('category', copy=False)

    # add nan as category
    if include_nan:
        nan_columns = data.columns[data.isna().any()]
        for c in nan_columns:
            data[c] = data[c].cat.add_categories('nan').fillna('nan')
    return data